#/***********************************************************************
#                          ARSP.pm  -  description
#                          ---------------------------
#    begin                : Wed Jun 13 2004
#    update               : Fri Aug 27 2004 
#    copyright            : (C) 2004 by Georg P. Loczewski
#    email                : gpl@lambda-bound.de
# **********************************************************************/
#
#/***********************************************************************
#*                                                                      *
#* This program is free software; you can redistribute it and/or modify *
#* it under the terms of the GNU General Public License as published by *
#* the Free Software Foundation; either version 2 of the License, or    *
#* (at your option) any later version.                                  *
#*                                                                      *
#***********************************************************************/

package Aplpl::ARSP;
use strict;

use Aplpl::NAME;
use Aplpl::VALUE;
use Aplpl::EXP;
use Aplpl::PAIR;
use Aplpl::ENV;

use Exporter;
use vars qw(@ISA @EXPORT);
@ISA = ("Exporter");
@EXPORT = qw(&newARSP);


sub newARSP {
   my $type   = 'ARSP';
   my @cbuffer = ();               # main input buffer 
   my $current = 0;		# current buffer pointer
   my $currmax = 0;;		# size of cbuffer

   my $skipSpaces = sub {};
   my $readEXP    = sub {};
   my $readNL     = sub {};
   my $readEL     = sub {};
   my $readQuote  = sub {};
   my $readNumber = sub {};
   my $readString = sub {};
   my $readSymbol = sub {};
   my $readNL_aux = sub {};
   my $readName   = sub {};

   my $parserInit = sub {
     my $ins = shift;

     @cbuffer = split //, $ins;

     $currmax = @cbuffer;
     # initialize the current pointer
     $current = 0;
     $skipSpaces->();
   };

   #/* ignore white space and comments */
   $skipSpaces = sub {
     if ($current < $currmax) {
        my $temp = $cbuffer[$current];
        while ($current < $currmax && $temp =~ m/[\s]/) {
          $current++;
          $temp = $cbuffer[$current];
        }
        if ($current < $currmax && $temp eq ';') {	 
           $current++;
           $temp = $cbuffer[$current];
           while ($current < $currmax && $temp =~ m/[^\n]/ ) {
              $current++;
              $temp = $cbuffer[$current];
           }
           $skipSpaces->();
        }
     }
   };
   
   #/* test for separator between tokens */
   my $isSeparator = sub {
     $_ = shift;
     if ($current < $currmax) {
        if (m/[\s'\";\)\(]/){      ## for print purposes: '
          return 1;
        }
        else {
           return 0;
        }
     }
     else {
        return 1;
     }
   };
   
   #/* read and parse an ars expression */
   $readEXP = sub {
     my $etemp;
     my $nm;
     my $nl;
     my $el;
     my $DEFINE = defName("define");
     my $LAMBDA = defName("lambda");
   
     $skipSpaces->();
     if ($current >= $currmax) {
        return $eof;
     }
     $_ = $cbuffer[$current];
     if (m/[(]/) {
       my $tag;
       $current++;
       $etemp = $readEXP->();
       if (($etemp->('etype') eq 'RXP') && 
           ($etemp->('evalue')->('value') == $LAMBDA)) {
         $tag = $LAMBDA;
         $nl  = $readNL->();
         $el  = $readEL->();
         return newEXP ('AXP', make_list(newEXP('VXP', $tag), $nl, $el));
       }
       elsif (($etemp->('etype') eq 'RXP') && 
              ($etemp->('evalue')->('value') == $DEFINE)) {
         $tag = $DEFINE;
         $el  = $readEL->();
         return newEXP ('AXP', elcons(newEXP('VXP', $tag), $el));
       }
       else {
         return newEXP ('SXP', elcons($etemp, $readEL->()));
       }
     }
     elsif (m/["]/) {               ## for print purposes: "              
       $current++;
        return newEXP ('VXP', $readString->());
     }
     elsif (m/[']/) {               ## for print purposes: '              
       $current++;
       return $readQuote->();
     }
     elsif (m/[0-9]/) {
        #$main::ots = $main::ots . sprintf("%s", "VXP: number: " . $_ . "\n");
        return newEXP('VXP', $readNumber->());
     }
     elsif (m/[A-Za-z_.\+\-\*\/\>\<\=]/) {
        #$main::ots = $main::ots . sprintf("%s", "RXP: symbol: " . $_ . "\n");
        return  newEXP('RXP', $readSymbol->());
     }
     else {
        print "abort:-" . $cbuffer[$current] . "-<br>";
        $main::ots = $main::ots . 
                     sprintf("%s", "parse error: invalid expression: $_ \n");
        die "Abnormal Termination.";
     }
   };
   
   #/* read quoted expression */
   $readQuote = sub {
     $current++;
     if ($cbuffer[$current] eq '(') {
        $current++;
        return newEXP('QLX', $readEL->());
     }
     else {
        return newEXP('QSX', $readSymbol->());
     }
   };
   
   #/* read and parse a list of expressions */
   $readEL = sub {
     my $etemp;
   
     $skipSpaces->();
     if ($cbuffer[$current] eq ')') {
        $current++;
        return $nil;
     }
     $etemp = $readEXP->();
     return elcons ($etemp, $readEL->());
   };
   
   #/* read and parse a list of names */
   $readNL = sub {
     my $nl;
     $skipSpaces->();
     my $cchar = $cbuffer[$current];
   
     if ($cchar eq '(') {
        $current++;
        $nl = $readNL_aux->();
        return $nl;
     }
     else {
        $main::ots = $main::ots . 
                     sprintf ("readNL: '(' expected: got %s\n", $cchar);
        die "Abnormal Termination.";
     }
   };
   
   #/* auxiliary function for readNL */
   $readNL_aux = sub {
     my $nm;
   
     $skipSpaces->();
     if ($cbuffer[$current] eq ')') {
        $current++;
        return $nil;
     }
     $nm = $readName->();
     return nlcons ($nm, $readNL_aux->());
   };
   
   #/* read and parse the name of a variable */
   $readName = sub {
     my @strbuf;
     my $nm;
   
     for (my $idx=0; !$isSeparator->($cbuffer[$current]); 
                                     $current++, $idx++) {
        $strbuf[$idx]= $cbuffer[$current];
     };
     $nm =  defName( join('',@strbuf) );
     return $nm;
   };
   
   #/* read and parse the name of a variable */
   $readSymbol = sub {
     my @strbuf;
     my $nm;
   
     for (my $idx=0; !$isSeparator->($cbuffer[$current]); 
                                              $current++, $idx++) {
        $strbuf[$idx]= $cbuffer[$current];
     };
     $nm =  defName( join('',@strbuf) );
     return newVAL('SYM', $nm);
   };
   
   #/* read and parse a number */
   $readNumber = sub {
     my @strbuf;
     my $numstr;
   
     for (my $idx=0; !$isSeparator->($cbuffer[$current]); 
                                              $current++, $idx++) {
        $strbuf[$idx]= $cbuffer[$current];
     };
     $numstr = join('',@strbuf);
     $_ = $numstr;
     if ($numstr =~ m/^0[xX][0-9a-fA-F]+$/) {
        return newVAL('INT', hex($numstr));
     }
     elsif ($numstr =~ m/^\d+$/) {
        return newVAL('INT', $numstr + 0);
     }
     elsif ($numstr =~ m/^[0-9]+\.[0-9]+$/) {
        return newVAL('DBL', $numstr + 0.0);
     }
     else {
        $main::ots = $main::ots . 
                     sprintf ("invalid number: %s\n", $numstr);
        die "Abnormal Termination.";
     }
   };
   
   #/* read and parse a string (identified by double quotes) */
   $readString = sub {
     my @strbuf;
     my $res = "";
     my $eos = 0;
   
     for (my $idx=0; !$eos; $current++, $idx++) {
        if ($cbuffer[$current] eq "\\") {    ## for print purposes: "              
           $current++;
        }
        $strbuf[$idx]= $cbuffer[$current];
        if ($cbuffer[$current+1] eq '"') {   ## for print purposes: "              
           $current++;
           $eos = 1;
        }
     }
     $res = join ('',@strbuf);
     return newVAL('STR', $res);
   };

   my $objref = sub {
                   my $msg = shift;
                   if ($msg eq 'type') {
                      return $type;
                   } 
                   elsif ($msg eq 'init') {
                      my $arg = shift;
                      return $parserInit->($arg);
                   } 
                   elsif ($msg eq 'readexp') {
                      return $readEXP->();
                   } 
                   else {
                      $main::ots = $main::ots . 
                                   "ARSP: message unknown: $msg \n";
                   }
                };
   return (bless($objref, 'Aplpl::ARSP'));
}
1;

