package Aplpl::EVAL;
use strict;
use subs qw(&parserInit);

use Aplpl::VALUE;
use Aplpl::PAIR;
use Aplpl::ENV;
use Aplpl::NAME;
use Aplpl::ARSP;
use Exporter;
if ($main::WWW) {
   use LWP::Simple;
}
use Exporter;
use vars qw(@ISA @EXPORT);
@ISA = ("Exporter");
@EXPORT = qw(&leval &evalList &evalThunk &evalSequence &applyARSClosure 
             &applyPrimitive &applySpecialForm &apply &load &prValue);


my $DEFINE = defName("define");
my $LAMBDA = defName("lambda");

sub leval {
   my $e    = shift;
   my $lenv = shift;
   my $lazy = shift;
  
   my $op;
   my $primname;
   my $result;
   my $temp;

  if ($e->('etype') eq 'VXP') {
      return $e->('evalue');
   }
   elsif ($e->('etype') eq 'QSX') {
      $result = $e->('evalue');
   }
   elsif ($e->('etype') eq 'QLX') {
      $result = newVAL('PAIR', $e->('evalue'));
   }
   elsif ($e->('etype') eq 'RXP') {
      if (isbound($e->('evalue')->('value'), $lenv)) {
         $result =  getVar($e->('evalue')->('value'), $lenv);
      }
      else {
         $main::ots = $main::ots . sprintf("%s", "Undefined variable: ");
         $main::ots = $main::ots . prName ($e->('evalue')->('value'));
         $main::ots = $main::ots . sprintf("%s", "\n");
         die "Abnormal Termination ";
      }
      $result = evalThunk($result);
   }
   elsif ($e->('etype') eq 'SXP') {
      my $result;
      $op = evalThunk (leval (car($e->('evalue')), $lenv, $lazy));
      if ($op->('vtype') eq 'PRIM') {
         $primname = $op->('value');
         return applyPrimitive ($primname, 
                                evalList(cdr($e->('evalue')), $lenv, $lazy));
      }
      else {
         return applyARSClosure ($op, 
                                 evalList(cdr($e->('evalue')), $lenv, $lazy),
                                 $lazy);
      }
   }
   elsif ($e->('etype') eq 'AXP') {
      if (car($e->('evalue'))->('evalue') == $DEFINE) {
         return applySpecialForm ($DEFINE,  cdr($e->('evalue')), $lenv);
      }
      elsif (car($e->('evalue'))->('evalue') == $LAMBDA) {
         return newVAL('ACL', cdr($e->('evalue')), $lenv);
      }
      else {
         $main::ots = $main::ots . 
                      sprintf("%s", 
                      "undefined expression type: $e->('etype')");
         die "Abnormal Termination.";
      }
   }
   else {
      $main::ots = $main::ots . 
                   sprintf("%s", 
                   "undefined expression type: $e->('etype')");
      die "Abnormal Termination.";
   } 
   return $result;
};


sub evalThunk {
   my $val = shift;

   if ($val->('type') eq 'VAL') {
      if ($val->('vtype') eq 'THUNK') {
         my $temp = $val->('value');
         if (car($temp)->('type') eq 'EXP') {
            return evalThunk (leval (car($temp), cdr($temp), $trueSym));
         }
         else {
            print "evalThunk: wrong type of sub-parameter!<br>";
            return $falseValue;
         }
      }
      else {
         return $val;
      }
   }
   else {
      print "evalThunk: wrong type of parameter!<br>";
      return $val;
   }
};

sub evalList {
   my $el   = shift;
   my $lenv = shift;
   my $lazy = shift;

   my $h;
   my $t;

   if ($el == $nil) {
      return $nil;
   }
   else {
      if (lazyp(car($el), $lazy)) {
         $h = newVAL('THUNK', car($el), $lenv);
      }
      else {
         $h = evalThunk(leval(car($el), $lenv, $lazy));
      }
      $t = evalList(cdr($el), $lenv, $lazy);
      return vlcons($h, $t);
   }
};

sub evalSequence {
   my $el     = shift;
   my $newenv = shift;
   my $lazy   = shift;

   while (cdr($el) != $nil) {
      evalThunk (leval (car($el), $newenv, $lazy));
      $el = cdr($el);
   } 
   return evalThunk (leval (car($el), $newenv, $lazy));
};

sub applyARSClosure {
   my $op   = shift;
   my $args = shift;
   my $lazy = shift;

   my $fun;
   my $body;
   my $names;
   my $savedenv;
   my $newenv;

   $fun      = car($op->('value'));
   $savedenv = cdr($op->('value'));
   $names    = car($fun);
   $body     = car(cdr($fun));
   if (llength($names) != llength($args)) {
      $main::ots = $main::ots . 
                   sprintf("%s", 
                   "Wrong number of arguments to closure\n");
      $main::ots = $main::ots . sprintf("%s", "names:\n");
      $main::ots = $main::ots . sprintf("%s", "\narguments:\n");
      die "Abnormal Termination";
   }
   else {
      $newenv = extendEnv ($savedenv, $names, $args);
      return evalSequence($body, $newenv, $lazy);
   }
};

sub applyPrimitive {
   my $op = shift;
   my $vl = shift;

   my $result;
   my $v1;
   my $v2;
   my $nargs = llength($vl);

   if ($nargs == 0) {
      if ($op == $PQUIT) {
         die "quit";
      }
      else {
         $main::ots = $main::ots . sprintf("%s", "illegal operation!");
         die "Abnormal Termination.";
      }
   }
   elsif ($nargs == 1) {
      $v1 = evalThunk (car($vl));
      if ($op == $PINCR) {
         if ($v1->('vtype') eq 'INT') {
            return newVAL('INT', $v1->('value') + 1);
         }
         else {
            $main::ots = $main::ots . 
                         sprintf("%s", "Integer argument expected!");
            prName ($op);
            $main::ots = $main::ots . sprintf("%s", "\n");
            die "Abnormal Termination";
         }
      }
      elsif ($op == $PPRINT) {
         $main::ots = $main::ots . sprintf("%s", "-->");
         prValue ($v1);
         $main::ots = $main::ots . sprintf("%s", "\n");
         return newVAL('SYM', defName("void\n"));
      }
      elsif ($op == $PLOAD) {
         return load ($v1);
      }
      else {
         print "undefined primitive operation!\n";
         prName ($op);
         print "\n";
         die "Abnormal Termination.";
      }
   } 
   elsif ($nargs == 2) {
      $v1 = evalThunk (car($vl));
      $v2 = evalThunk (car(cdr($vl)));
      if ($op == $PADD) {
         if (($v1->('vtype') eq 'INT') && ($v2->('vtype') eq 'INT')) {
            $result =  newVAL('INT', $v1->('value') + $v2->('value'));
            return $result;
         }
         else {
            $main::ots = $main::ots . 
                         sprintf("%s\n", "Integer argument expected!");
            die "Abnormal Termination";
         }
      }
      elsif ($op == $PSUB) {
         if (($v1->('vtype') eq 'INT') && ($v2->('vtype') eq 'INT')) {
            $result =  newVAL('INT', $v1->('value') - $v2->('value'));
            return $result;
         }
         else {
            $main::ots = $main::ots . 
                         sprintf("%s\n", "Integer argument expected!");
            die "Abnormal Termination";
         }
      }
      elsif ($op == $PMLT) {
         if (($v1->('vtype') eq 'INT') && ($v2->('vtype') eq 'INT')) {
            $result =  newVAL('INT', $v1->('value') * $v2->('value'));
            return $result;
         }
         else {
            $main::ots = $main::ots . 
                         sprintf("%s\n", "Integer argument expected!");
            die "Abnormal Termination";
         }
      }
      elsif ($op == $PDIV) {
         if (($v1->('vtype') eq 'INT') && ($v2->('vtype') eq 'INT')) {
            $result =  newVAL('INT', $v1->('value') / $v2->('value'));
            return $result;
         }
         else {
            $main::ots = $main::ots . 
                         sprintf("%s\n", "Integer argument expected!");
            die "Abnormal Termination";
         }
      }
      elsif ($op == $PGE) {
         if ($v1->('vtype') ne $v2->('vtype')) {
            $result = getVar(defName("false"), $main::globalEnv);
         }
         elsif ($v1->('vtype') eq 'INT' &&
                $v1->('value') >= $v2->('value')) {
            $result = getVar(defName("true"), $main::globalEnv);

         }
         elsif ($v1->('vtype') eq 'DBL' &&
                $v1->('value') >= $v2->('value')) {
            $result = getVar(defName("true"), $main::globalEnv);

         }
         else {
            $result = getVar(defName("false"), $main::globalEnv);
         }
      }
      elsif ($op == $PLT) {
         if ($v1->('vtype') ne $v2->('vtype')) {
            $result = getVar(defName("false"), $main::globalEnv);
         }
         elsif ($v1->('vtype') eq 'INT' &&
                $v1->('value') < $v2->('value')) {
            $result = getVar(defName("true"), $main::globalEnv);

         }
         elsif ($v1->('vtype') eq 'DBL' &&
                $v1->('value') < $v2->('value')) {
            $result = getVar(defName("true"), $main::globalEnv);

         }
         else {
            $result = getVar(defName("false"), $main::globalEnv);
         }
      }
      elsif ($op == $PEQUAL) {
         if ($v1 == $v2) {
            $result = getVar(defName("true"), $main::globalEnv);
         }
         elsif ($v1->('vtype') ne $v2->('vtype')) {
            $result = getVar(defName("false"), $main::globalEnv);
         }
         elsif ($v1->('vtype') eq 'INT' &&
                $v1->('value') == $v2->('value')) {
            $result = getVar(defName("true"), $main::globalEnv);

         }
         elsif ($v1->('vtype') eq 'DBL' &&
                $v1->('value') == $v2->('value')) {
            $result = getVar(defName("true"), $main::globalEnv);

         }
         elsif ($v1->('vtype') eq 'SYM') {
            if ($v1->('value') eq $v2->('value')) {
               $result = getVar(defName("true"), $main::globalEnv);
            }
            else {
               $result = getVar(defName("false"), $main::globalEnv);
            }
         }
         elsif ($v1->('vtype') eq 'STR') {
            if ($v1->('value') eq $v2->('value')) {
               $result = getVar(defName("true"), $main::globalEnv);
            }
            else {
               $result = getVar(defName("false"), $main::globalEnv);
            }
         }
         else {
            $result = getVar(defName("false"), $main::globalEnv);
         }
      }
      else {
         print "undefined primitive operation!\n";
         prName ($op);
         print "\n";
         die "Abnormal Termination.";
      }
      
   } 
   else {
      print "Wrong number of arguments to primitive!";
      prName ($op);
      print "\n";
      die "Abnormal Termination.";
   }
};

sub applySpecialForm {
   my $op   = shift;
   my $args = shift;
   my $lenv = shift;

   my $result;

   if ($op == $DEFINE) {
      $result = evalThunk (leval (car(cdr($args)), $lenv, car($args))); 
      if (car($args)->('etype') eq 'RXP') {
         if (isbound (car($args)->('evalue')->('value'), $lenv)) {
            setVar (car($args)->('evalue')->('value'), $result, $lenv);
         }
         else {
            defVar (car($args)->('evalue')->('value'), $result, $lenv);
         } 
      }
      else {
         $main::ots = $main::ots . 
                      sprintf("%s", "Define: symbol expected!\n");
         die "Abnormal Termination.";
      }
   }
   else {
      $main::ots = $main::ots . 
                   sprintf("%s", "undefined control operation!\n");
      die "Abnormal Termination.";
   }
   return $result;
};

sub lazyp {
   my $e1 = shift;
   my $e2 = shift;

   if ($e1->('etype') eq 'RXP' && $e2->('etype') eq 'RXP') {
      if ($e1->('evalue')->('value') == $e2->('evalue')->('value')) {
         return 0;
      }
      else {
         return 1;
      }
   }
   else {
      return 1;
   }
};

sub prValue {
   my $val = shift;
 
   if ($val->('vtype') eq 'INT') {
      $main::ots = $main::ots . sprintf("%d", $val->('value'));
   }
   elsif ($val->('vtype') eq 'DBL') {
      $main::ots = $main::ots . sprintf("%g", $val->('value'));
   }
   elsif ($val->('vtype') eq 'SYM') {
      $main::ots = $main::ots . getName($val->('value'));
   }
   elsif ($val->('vtype') eq 'STR') {
      $main::ots = $main::ots . sprintf("%s", $val->('value'));
   }
   elsif ($val->('vtype') eq 'PAIR') {
      my $i = 0;
      $main::ots = $main::ots . sprintf("%s","(");
      while ($val != $nil) {
         $i++;
         prValue (evalThunk (car($val->('value'))));
         print " ";
         $val = cdr($val->('value'));
      }
      $main::ots = $main::ots . sprintf("%s", ")\n");
   }
   elsif ($val->('vtype') eq 'PRIM') {
      $main::ots = $main::ots . sprintf("%s", "<primitive: ");
      prName($val->('value'));
      $main::ots = $main::ots . sprintf("%s", ">");
   }
   elsif ($val->('vtype') eq 'ACL') {
      $main::ots = $main::ots . sprintf("%s", "ACL: lambda");
      prNL(car(car($val->('value'))));
   }
   elsif ($val->('vtype') eq 'THUNK') {
      $main::ots = $main::ots . sprintf("%s", "...");
   }
   else {
      $main::ots = $main::ots . sprintf("%s", "undefined value!");
      die "Abnormal Termination.";
   }
   #$main::ots = $main::ots . sprintf("%s", "\n");
}

sub load {
   my $val = shift;

   my $parser;
   my $file;
   my $url;
   my $content = "";
   my $result;
   my $exp;

   $parser = newARSP();
   if ($main::WWW) {
      if ($val->('vtype') eq 'STR') {
         $url = $val->('value');
         unless (defined ($content = get $url)) {
                die "could not get $url\n";
         }
      }
      else {
         $main::ots = $main::ots . 
                      sprintf("%s", "load: error in argument type\n");
         die "Abnormal Termination.";
      }
      $parser->('init', $content .  "\"*eof*\"");
      $exp = $parser->('readexp');
      while ($exp != $eof) {
         $result = evalThunk(leval($exp, $main::globalEnv, $trueSym));
         prValue($result);
         $exp = $parser->('readexp');
      }
      return newVAL('STR', "loading of \"$url\" completed!\n");
   }
   else {
      if ($val->('vtype') eq 'STR') {
         $file = $val->('value');
         open (INS, "< $file");
         while(<INS>) {
            $content = $content . $_;
         }
         close INS;
      }
      else {
         $main::ots = $main::ots . 
                      sprintf("%s", "load: error in argument type\n");
         die "Abnormal Termination.";
      }
      print $main::ots . "\n";
      $main::ots = "";
      print "loading file: $file ...\n";
      $parser->('init', $content .  "\"*eof*\"");
      $exp = $parser->('readexp');
      while ($exp != $eof) {
         $result = evalThunk(leval($exp, $main::globalEnv, $trueSym));
         prValue($result);
         print $main::ots;
         $main::ots = "";
         $exp = $parser->('readexp');
      }
      return newVAL('STR', "loading of \"$file\" completed!\n");
   }
};

1;
